/* Utility functions related to gamma functions.  Use the algorithms in numerical recipes.
 */

# include <stdio.h>
# include <stdlib.h>
# include <string.h>
# include <math.h>
# include "cluster.h"
# include "integration5.h"
# define MAXITERATIONS 1000
# define EPS 3e-30
# define FPMIN 1e-30
# define INFINITY 1e10
# define pi 3.14159
# define MAXLINEWIDTH 1000

extern double atof2(char *s, int length);
extern char * getitemval(char *s, int ind, int *length);
extern int pair_cmp(struct pair *p1, struct pair *p2);
extern double mean(long N, double *x);
extern double cov(long N, double *x, double *y);
double norminv(double p);
double erfinv(double y);
double erf(double x);

char *chi2lookuptable="chi2cdf1table.txt";



// The chi2cdf function of degree one by checking the lookup table.
// mode=0: return the cdf, mode=1: retrun 1-cdf.
double chi2cdf_deg1(double x, int mode)
{
  int length, flag, cnt=0;
  double x1, x2, p1, p2, q1, q2, p=0, q=0;
  char *s, *item;
  FILE *fp;

  if (x<=0)
    {
      if (mode==0)
	return 0.0;
      else
	return 1.0;
    }
  else if (x>=70.0)
    {
      if (mode==0)
	return 1.0;
      else 
	return 0.0;
    }

  // Open the lookup table.
  if ((fp=fopen(chi2lookuptable,"r"))==NULL)
    {
      printf("Cannot open %s.\n",chi2lookuptable);
      return 0.0;
    }
  // Find the appropriate value of x.
  s=malloc(sizeof(char)*MAXLINEWIDTH);
  fgets(s,MAXLINEWIDTH,fp); fgets(s,MAXLINEWIDTH,fp);

  x1=0; p1=0; q1=1; flag=0; cnt=2;
  while ((fgets(s,MAXLINEWIDTH,fp)!=NULL)&&(flag==0))
    {
      cnt++;

      item=getitemval(s,0,&length);
      x2=atof2(item,length);
      free(item);
      item=getitemval(s,1,&length);
      p2=atof2(item,length);
      free(item);
      item=getitemval(s,2,&length);
      q2=atof2(item,length);
      free(item);
      
      if (x==x2)
	{
	  p=p2; q=q2; flag=1;
	}
      else if ((x<x2)&&(x>x1))
	{
	  p=(p1+p2)/2; q=(q1+q2)/2; flag=1;
	}
      x1=x2; p1=p2; q1=q2;
    }
  free(s); fclose(fp);
  
  if (mode==0)
    return p;
  else
    return q;
}


// The inverse chi2cdf function of degree one by checking the lookup table.
// mode=0: input cdf, mode=1: intput 1-cdf.
double invchi2cdf_deg1(double val, int mode)
{
  int length, flag, cnt=2;
  double x1, x2, p1, p2, q1, q2, x, p, q;
  char *s, *item;
  FILE *fp;

  if (val<=0)
    return 0.0;
  else if (val>=1.0)
    return 70.0;
  
  // Open the lookup table.
  if ((fp=fopen(chi2lookuptable,"r"))==NULL)
    {
      printf("Cannot open %s.\n",chi2lookuptable);
      return 0.0;
    }
  
  if (mode==0)
    {
      p=val; q=1-p;
    }
  else
    {
      q=val; p=1-q;
    }

  // Find the appropriate value of x.
  s=malloc(sizeof(char)*MAXLINEWIDTH);
  fgets(s,MAXLINEWIDTH,fp); fgets(s,MAXLINEWIDTH,fp);
  x1=0; p1=0; q1=1; flag=0; x=0;
  while ((fgets(s,MAXLINEWIDTH,fp)!=NULL)&&(flag==0))
    {
      cnt++;

      item=getitemval(s,0,&length);
      x2=atof2(item,length);
      free(item);
      item=getitemval(s,1,&length);
      p2=atof2(item,length);
      free(item);
      item=getitemval(s,2,&length);
      q2=atof2(item,length);
      free(item);

      if ((mode==0)&&(p==p2))
	{
	  x=x2; flag=1;
	}
      else if ((mode==1)&&(q==q2))
	{
	  x=x2; flag=1;
	}
      else if ((mode==0)&&(p<p2)&&(p>p1))
	{
	  x=(x1+x2)/2; flag=1;
	}
      else if ((mode==1)&&(q>q2)&&(q<q1))
	{
	  x=(x1+x2)/2; flag=1;
	}
      x1=x2; p1=p2; q1=q2;
    }
  free(s); fclose(fp);
  
  return x;
}


// The chi2cdf function of degree one by checking the lookup table.
// mode=0: return the cdf, mode=1: retrun 1-cdf.
// Difference from chi2cdf_deg1: apply binary search on the data structure.
double chi2cdf_deg1_2(double x, int mode, int Nentries, struct chi2cdf_lookup *chi2table)
{
  int flag=0, lowind, uppind, ind;
  double x1, x2, p1, p2, q1, q2, p, q;
  
  lowind=0; uppind=Nentries-1; p=0.0; q=1.0;
  x1=(chi2table+lowind)->x; x2=(chi2table+uppind)->x; 
  ind=(int)((lowind+uppind)/2);

  if (x<x1)
    {
      p=0.0; q=1.0;
      if (mode==0)
	return p;
      else
	return q;
    }
  else if (x>x2)
    {
      p=1.0; q=0.0;
      if (mode==0)
	return p;
      else
	return q;
    }

  while (flag==0)
    {
      // Debug
      //printf("lowind=%d, uppind=%d, ind=%d, x1=%f, x2=%f, x=%f.\n",lowind,uppind,ind,x1,x2,x);

      if (x==x1)
	{
	  flag=1; ind=lowind; p=(chi2table+ind)->p; q=(chi2table+ind)->q; 
	}
      else if (x==x2)
	{
	  flag=1; ind=uppind; p=(chi2table+ind)->p; q=(chi2table+ind)->q;
	}
      else if (x==(chi2table+ind)->x)
	{
	  flag=1; p=(chi2table+ind)->p; q=(chi2table+ind)->q;
	}
      else if ((x>x1)&&(x<x2)&&(lowind>=(uppind-1)))
	{
	  flag=1; 
	  //p=((chi2table+lowind)->p+(chi2table+uppind)->p)/2;
	  //q=((chi2table+lowind)->q+(chi2table+uppind)->q)/2;

	  // Interpolate p and q.
	  p1=(chi2table+lowind)->p; q1=(chi2table+lowind)->q;
	  p2=(chi2table+uppind)->p; q2=(chi2table+uppind)->q;
	  p=p1+(p2-p1)*(x-x1)/(x2-x1);
	  q=q1+(q2-q1)*(x-x1)/(x2-x1);
	}
      else if ((x>x1)&&(x<(chi2table+ind)->x))
	{
	  uppind=ind; ind=(int)(lowind+uppind)/2;
	  x1=(chi2table+lowind)->x; x2=(chi2table+uppind)->x;
	}
      else if ((x<x2)&&(x>(chi2table+ind)->x))
	{
	  lowind=ind; ind=(int)(lowind+uppind)/2;
	  x1=(chi2table+lowind)->x; x2=(chi2table+uppind)->x;
	}

      // Debug
      //if (flag==1)
      //printf("flag=1, lowind=%d, uppind=%d, ind=%d, x1=%f, x2=%f, x=%f, p=%e, q=%e.\n",lowind,uppind,ind,x1,x2,x,p,q);

    }
      
  if (mode==0)
    return p;
  else
    return q;
}


// The inverse chi2cdf function of degree one by checking the lookup table.
// mode=0: return the cdf, mode=1: retrun 1-cdf.
// Difference from invchi2cdf_deg1: apply binary search on the data structure.
double invchi2cdf_deg1_2(double val, int mode, int Nentries, struct chi2cdf_lookup *chi2table)
{
  int flag=0, lowind, uppind, ind;
  double x1, x2, p1, p2, q1, q2, p, q, x=0;
  
  if (val<=0)
    return 0.0;
  else if (val>=1.0)
    return (chi2table+Nentries-1)->x;
    
  if (mode==0)
    {
      p=val; q=1-p;
    }
  else
    {
      q=val; p=1-q;
    }

  lowind=0; uppind=Nentries-1; ind=(int)((Nentries-1)/2); flag=0;
  p1=(chi2table+lowind)->p; q1=(chi2table+lowind)->q;
  p2=(chi2table+uppind)->p; q2=(chi2table+uppind)->q;

  if ((mode==0)&&(val<p1))
    return (chi2table+lowind)->x;
  else if ((mode==0)&&(val>p2))
    return (chi2table+uppind)->x;
  else if ((mode==1)&&(val>q1))
    return (chi2table+lowind)->x;
  else if ((mode==1)&&(val<q2))
    return (chi2table+uppind)->x;


  while (flag==0)
    {
      if ((mode==0)&&(val==p1))
	{
	  flag=1; ind=lowind; x=(chi2table+ind)->x;
	}
      else if ((mode==1)&&(val==q1))
	{
	  flag=1; ind=lowind; x=(chi2table+ind)->x;
	}
      else if ((mode==0)&&(val==p2))
	{
	  flag=1; ind=uppind; x=(chi2table+ind)->x;
	}
      else if ((mode==1)&&(val==q2))
	{
	  flag=1; ind=uppind; x=(chi2table+ind)->x;
	}
      else if ((mode==0)&&(val==(chi2table+ind)->p))
	{
	  flag=1; x=(chi2table+ind)->x;
	}
      else if ((mode==1)&&(val==(chi2table+ind)->q))
	{
	  flag=1; x=(chi2table+ind)->x;
	}
      else if ((mode==0)&&(val>p1)&&(val<p2)&&(lowind>=(uppind-1)))
	{
	  flag=1; // x=((chi2table+lowind)->x+(chi2table+uppind)->x)/2;
	  
	  // Interpolate x.
	  x1=(chi2table+lowind)->x; x2=(chi2table+uppind)->x;
	  x=x1+(x2-x1)*(val-p1)/(p2-p1);
	}
      else if ((mode==1)&&(val<q1)&&(val>q2)&&(lowind>=(uppind-1)))
	{
	  flag=1; // x=((chi2table+lowind)->x+(chi2table+uppind)->x)/2;

	  // Interpolate x.
	  x1=(chi2table+lowind)->x; x2=(chi2table+uppind)->x;
	  x=x1+(x2-x1)*(val-q1)/(q2-q1);
	}
      else if ((mode==0)&&(val>p1)&&(val<(chi2table+ind)->p))
	{
	  uppind=ind; ind=(int)(lowind+uppind)/2;
	  p1=(chi2table+lowind)->p; q1=(chi2table+lowind)->q;
	  p2=(chi2table+uppind)->p; q2=(chi2table+uppind)->q;
	}
      else if ((mode==1)&&(val<q1)&&(val>(chi2table+ind)->q))
	{
	  uppind=ind; ind=(int)(lowind+uppind)/2;
	  p1=(chi2table+lowind)->p; q1=(chi2table+lowind)->q;
	  p2=(chi2table+uppind)->p; q2=(chi2table+uppind)->q;
	}
      else if ((mode==0)&&(val<p2)&&(val>(chi2table+ind)->p))
	{
	  lowind=ind; ind=(int)(lowind+uppind)/2;
	  p1=(chi2table+lowind)->p; q1=(chi2table+lowind)->q;
	  p2=(chi2table+uppind)->p; q2=(chi2table+uppind)->q;
	}
      else if ((mode==1)&&(val>q2)&&(val<(chi2table+ind)->q))
	{
	  lowind=ind; ind=(int)(lowind+uppind)/2;
	  p1=(chi2table+lowind)->p; q1=(chi2table+lowind)->q;
	  p2=(chi2table+uppind)->p; q2=(chi2table+uppind)->q;
	}
    }

  return x;
}
