
// A composite edge containing three categories of edges.
struct cedge {
  int type;  // 0:regulated edges, 1:signal edges, 2:indirect edges.
  int from;  // source of the edge, no directionality if the edge is type 2.
  int to;    // destination of the edge, no directionality if the edge is type 2.
  int sign;  // the sign of edge interaction.
  //  int dir;   // the direction of a signal edge.
  double priorp; // the prior probability of edge existence, conditioned on single location/pp p-value at that edge.
  double postp; // the posterior probability of edge existence conditioned on the global evidence
};

// A sparse graph containing composite edges.
struct csparsegraph {
  int Nv;  // number of nodes.
  int *v;  // nodes.
  int Ne;  // number of edges.
  struct cedge *e;  // edges.
  struct clique *nodeclus; // clusters associated with each node.
};


// A path containing composite edges and is annotated by a sign.
struct cpath {
  int Ne;  // number of edges.
  int *e;  // edge indices in the graph.
  int sign;    // the sign of a path.
};

// A compressed cpath.
struct compresscpath {
  int start;   // start node.
  int pathid;  // the index of the intermediate path.
  int end;     // end node.
  int lastdir; // the direction of the last edge along the path, +1 if point toward the end node, -1 is point away from the end node.
  int length;  // the length of the path.
  int lastsign;    // the sign of the last edge in the path.
  int lasttype;  // the type of the last edge.
};


// A number of path sequences.
struct pathseqs {
  int Nseqs;  // number of path sequences.
  struct clique *seqs;  // path sequences; the element in each clique denotes the index of the path.
};


// A concatenation tree data structure.
struct concatenationnode {
  int curpos;     // current position.
  int curpathid;  // current path id.
  int parent;     // parent.
  int Nchildren;  // number of children.
  int *children;  // children.
};


// Probability table of binary random variables.
struct binprobtab {
  int Nvars;      // number of variables.
  int *varinds;   // variable indices.
  int Nconfigs;   // number of configurations.
  double *probs;  // probability of each configuration.
};


// A sparsegraph of junction tree.  Each node contains two arrays of message queues recording where messages come from and go to.

struct mnode {
  int ind;
  int Nmsgfrom;
  int *msgfrom;
  int Nmsgto;
  int *msgto;
};

struct msparsegraph {
  int Nv;           // number of nodes.
  struct mnode *v;  // nodes.
  int Ne;           // number of edges.
  struct cedge *e;  // edges.
};



// A general potential function between an arbitrary number of variables, where each variable has an arbitrary number of states.
struct potential {
  int Nvars;          // numebr of variables.
  int *varinds;          // variable indices.
  int Nconfigs;          // total number of configurations.
  int *Nstates;          // number of states in each variable.
  double *probs;         // probability of each configuration.
};


// A Markov random field graph.
struct MRF {
  int Nv;          // number of nodes.
  int *v;          // nodes.
  int Ne;          // number of edges.
  struct edge *e;  // edges (undirected).
  struct potential *phiv;  // potential functions of nodes.
  struct potential *phie;  // potential functions of edges.
};


// A factor graph data structure.  A bipartite graph of variable and factor nodes.
struct facgraph {
  int Nvars;          // number of variable nodes.
  int *vars;          // variable nodes.
  int Ninteractions;  // number of factor (interaction) nodes.
  int *interactions;  // interaction nodes.
  int Ne;             // number of edges.
  struct edge *e;     // edges (from variable nodes to factor nodes).
  struct potential *phivi;  // potential functions of factor nodes.
};


// A message data structure.
struct message {
  int from;  // from node.
  int to;    // to node.
  int dir;   // message direction, 0:variable to factor, 1:factor to variable.
  struct potential *phi;  // potential function of the message.
};


// The tree structure of variable configurations.
struct confignode {
  int ind;        // node index.
  int level;      // level of the node.
  int Nvars;      // number of variables.
  int *config;    // configuration.
  int Nchildren;  // number of children nodes.
  struct confignode *children; // children node.
};
